#include "Python.h"

#include "acpica.h"
#include "smp.h"

static PyObject *bits_cpus(PyObject *self, PyObject *args)
{
    int ncpus;
    int ndx;
    PyObject *apicid_list;
    const CPU_INFO *cpu;

    if (!PyArg_ParseTuple(args, ""))
        return NULL;
    ncpus = smp_init();
    apicid_list = PyList_New(ncpus);
    cpu = smp_read_cpu_list();
    for (ndx = 0; ndx < ncpus; ndx++)
        PyList_SET_ITEM(apicid_list, ndx, Py_BuildValue("i", cpu[ndx].apicid));
    return Py_BuildValue("O", apicid_list);
}

struct msr {
    U32 num;
    U32 status;
    U64 value;
};

static void msr_callback(void *param)
{
    struct msr *m = param;
    rdmsr64(m->num, &m->value, &m->status);
}

static PyObject *bits_rdmsr(PyObject *self, PyObject *args)
{
    struct msr msr;
    unsigned apicid;
    if (!PyArg_ParseTuple(args, "II", &apicid, &msr.num))
        return NULL;
    if (!smp_function(apicid, msr_callback, &msr))
        return Py_BuildValue("");
    if (msr.status)
        return Py_BuildValue("");
    return Py_BuildValue("K", msr.value);
}

static PyObject *acpi_object_to_python(ACPI_OBJECT *obj)
{
    switch(obj->Type) {
    case ACPI_TYPE_INTEGER:
        return Py_BuildValue("K", obj->Integer.Value);
    case ACPI_TYPE_STRING:
        return Py_BuildValue("s#", obj->String.Pointer, obj->String.Length);
    case ACPI_TYPE_BUFFER:
        return Py_BuildValue("s#", obj->Buffer.Pointer, obj->Buffer.Length);
    case ACPI_TYPE_PACKAGE:
        {
            U32 ndx;
            PyObject *pkg = PyTuple_New(obj->Package.Count);
            for (ndx = 0; ndx < obj->Package.Count; ndx++)
                PyTuple_SET_ITEM(pkg, ndx, acpi_object_to_python(&obj->Package.Elements[ndx]));
            return Py_BuildValue("O", pkg);
        }
    case ACPI_TYPE_PROCESSOR:
        return Py_BuildValue("III", obj->Processor.ProcId, obj->Processor.PblkAddress, obj->Processor.PblkLength);
    default:
        grub_printf("Internal error: acpi_object_to_python on unhandled ACPI_OBJECT_TYPE %u\n", obj->Type);
        return Py_BuildValue("");
    }
}

static PyObject *bits_acpi_eval(PyObject *self, PyObject *args)
{
    char *pathname;
    ACPI_BUFFER results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    PyObject *ret;

    if (!PyArg_ParseTuple(args, "s", &pathname))
        return NULL;

    if (acpica_init() != GRUB_ERR_NONE)
        return NULL;

    if (ACPI_FAILURE(AcpiEvaluateObject(NULL, pathname, NULL, &results)))
        return Py_BuildValue("");

    ret = acpi_object_to_python(results.Pointer);

    ACPI_FREE(results.Pointer);
    return Py_BuildValue("O", ret);
}

static PyObject *bits_acpi_get_table(PyObject *self, PyObject *args)
{
    char *signature;
    U32 instance = 0;
    ACPI_TABLE_HEADER *table_header;

    if (!PyArg_ParseTuple(args, "s|I", &signature, instance))
        return NULL;

    if (acpica_init() != GRUB_ERR_NONE)
        return NULL;

    if (ACPI_FAILURE(AcpiGetTable(signature, instance, &table_header)))
        return Py_BuildValue("");

    return Py_BuildValue("s#", table_header, table_header->Length);
}

struct find_processor_context {
    bool init_cpu;
    U32 caps;
    PyObject *cpupath_list;
};

static grub_err_t call_osc(ACPI_HANDLE cpu_handle, U32 caps)
{
    grub_err_t ret;
    ACPI_STATUS Status;
    ACPI_BUFFER Results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    ACPI_OBJECT_LIST Params;
    ACPI_OBJECT Obj[4];
    U32 osc_buffer[2];
    U8 OSC_UUID[16] = {0x16, 0xA6, 0x77, 0x40, 0x0C, 0x29, 0xBE, 0x47, 0x9E, 0xBD, 0xD8, 0x70, 0x58, 0x71, 0x39, 0x53};

    /* Initialize the parameter list */
    Params.Count = 4;
    Params.Pointer = Obj;

    /* Initialize the parameter objects */
    // Intel-specific UUID
    Obj[0].Type = ACPI_TYPE_BUFFER;
    Obj[0].Buffer.Length = 16;
    Obj[0].Buffer.Pointer = OSC_UUID;

    // Revision ID
    Obj[1].Type = ACPI_TYPE_INTEGER;
    Obj[1].Integer.Value = 1;

    // Count of DWORDS
    Obj[2].Type = ACPI_TYPE_INTEGER;
    Obj[2].Integer.Value = 2;

    // Capabilities buffer
    Obj[3].Type = ACPI_TYPE_BUFFER;
    Obj[3].Buffer.Length = 8;
    Obj[3].Buffer.Pointer = (U8 *)osc_buffer;

    osc_buffer[0] = 0;
    osc_buffer[1] = caps; // Capabilities DWORD

    Status = AcpiEvaluateObject(cpu_handle, "_OSC", &Params, &Results);
    if (Status == AE_OK)
        ret = GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND)
        ret = grub_error(GRUB_ERR_TEST_FAILURE, "false");
    else
        ret = grub_error(GRUB_ERR_IO, "Evaluating _OSC failed (0x%x %s)\n", Status, AcpiFormatException(Status));

    ACPI_FREE(Results.Pointer);

    return ret;
}

static grub_err_t call_pdc(ACPI_HANDLE cpu_handle, U32 caps)
{
    ACPI_STATUS Status;
    ACPI_OBJECT_LIST Params;
    ACPI_OBJECT Obj;
    U32 pdc_buffer[3];

    /* Initialize the parameter list */
    Params.Count = 1;
    Params.Pointer = &Obj;

    /* Initialize the parameter objects */
    Obj.Type = ACPI_TYPE_BUFFER;
    Obj.Buffer.Length = 12;
    Obj.Buffer.Pointer = (UINT8 *)pdc_buffer;

    /* Initialize the DWORD array */
    pdc_buffer[0] = 1; // Revision
    pdc_buffer[1] = 1; // Count
    pdc_buffer[2] = caps; // Capabilities DWORD

    Status = AcpiEvaluateObject(cpu_handle, "_PDC", &Params, NULL);
    if (Status == AE_OK)
        return GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND)
        return grub_error(GRUB_ERR_TEST_FAILURE, "false");
    else
        return grub_error(GRUB_ERR_IO, "Evaluating _PDC failed (0x%x %s)\n", Status, AcpiFormatException(Status));
}

static ACPI_STATUS find_processor(ACPI_HANDLE ObjHandle, UINT32 NestingLevel ACPI_UNUSED_VAR, void *Context, void **ReturnValue ACPI_UNUSED_VAR)
{
    struct find_processor_context *fpc = Context;
    ACPI_BUFFER Path = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    PyObject *cpupath;

    if (!IsEnabledProcessor(ObjHandle))
        goto out;

    if (ACPI_FAILURE(AcpiGetName(ObjHandle, ACPI_FULL_PATHNAME, &Path))) {
        grub_printf("Couldn't get object name\n");
        goto out;
    }

    if (fpc->init_cpu)
        if (call_osc(ObjHandle, fpc->caps) != GRUB_ERR_NONE)
            call_pdc(ObjHandle, fpc->caps);

    cpupath = Py_BuildValue("s", Path.Pointer);
    PyList_Append(fpc->cpupath_list, cpupath);
    Py_XDECREF(cpupath);

out:
    ACPI_FREE(Path.Pointer);
    return AE_OK;
}

static PyObject *bits_acpi_cpupaths(PyObject *self, PyObject *args)
{
    struct find_processor_context fpc = { .init_cpu = false, .caps = 0xfbf, .cpupath_list = NULL };

    if (!PyArg_ParseTuple(args, ""))
        return NULL;

    if (acpica_init() != GRUB_ERR_NONE)
        return NULL;

    if (acpica_cpus_initialized) {
        if (fpc.caps != acpica_cpus_init_caps) {
            grub_printf("CPUs already initialized with caps 0x%x; cannot reinit with 0x%x; call acpi_terminate first", acpica_cpus_init_caps, fpc.caps);
            return NULL;
        }
    } else {
        acpica_cpus_initialized = true;
        acpica_cpus_init_caps = fpc.caps;
        fpc.init_cpu = true;
    }

    fpc.cpupath_list = PyList_New(0);

    AcpiWalkNamespace(ACPI_TYPE_PROCESSOR, ACPI_ROOT_OBJECT, ACPI_UINT32_MAX, find_processor, NULL, &fpc, NULL);

    return fpc.cpupath_list;
}

static PyMethodDef bitsMethods[] = {
    {"acpi_cpupaths",  bits_acpi_cpupaths, METH_VARARGS, "acpi_cpupaths() -> list of cpu namepaths"},
    {"acpi_eval",  bits_acpi_eval, METH_VARARGS, "acpi_eval(\"\\PATH._TO_.EVAL\") -> result"},
    {"acpi_get_table",  bits_acpi_get_table, METH_VARARGS, "acpi_get_table(signature[, instance]) -> str"},
    {"cpus",  bits_cpus, METH_VARARGS, "cpus() -> list of APIC IDs"},
    {"rdmsr",  bits_rdmsr, METH_VARARGS, "rdmsr(apicid, msr) -> long (None if GPF)"},
    {NULL, NULL, 0, NULL}        /* Sentinel */
};

PyMODINIT_FUNC initbits(void)
{
    (void) Py_InitModule("bits", bitsMethods);
}


